#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import shutil

from pkg_resources import Requirement, cleanup_resources, resource_filename


_REQUIREMENT = Requirement.parse("PALEOMIX")


def rscript(tool, script):
    """Returns the path to an Rscript for a given tool."""

    path = os.path.join("paleomix", "resources", "rscripts", tool, script)

    return resource_filename(_REQUIREMENT, path)


def report(tool, filename):
    """Returns the path to a report-file for a given tool."""

    path = os.path.join("paleomix", "resources", "reports", tool, filename)

    return resource_filename(_REQUIREMENT, path)


def add_copy_example_command(subparsers):
    parser = subparsers.add_parser("example", help="Create example project")

    parser.add_argument(
        "destination",
        default=".",
        nargs="?",
        help="Destination folder for example data.",
    )


def copy_example(tool, args):
    """Command-line interface to copy a folder containing example data to a
    folder specified by the user. Arguments are a tool name (e.g.
    'bam_pipeline'), and any command-line options specified by the user;
    returns 0 on success, or 1 on errors.
    """
    log = logging.getLogger(__name__)
    destination = os.path.join(args.destination, tool)
    log.info("Copying example project to %r", args.destination)

    if os.path.exists(destination):
        log.error("Example folder already exists at %r", destination)
        return 1

    try:
        source = resource_filename("paleomix.resources", os.path.join("examples", tool))

        shutil.copytree(source, destination)
    finally:
        cleanup_resources()

    log.info("Sucessfully saved example in %r", destination)

    return 0



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix
import paleomix.resources

from paleomix.node import Node

import paleomix.common.fileutils as fileutils

import paleomix.pipelines.zonkey.parts.admixture as admixture

from paleomix.pipelines.zonkey.parts.report import AnalysisReport


class SummaryNode(Node):
    def __init__(self, config, dependencies=()):
        self._root = config.destination
        self._data = config.database
        self._samples = config.samples
        self._sample_keys = list(self._samples.keys())

        input_files = set()
        self._reports = {}
        for sample, info in self._samples.items():
            report = AnalysisReport(
                config=config,
                root=os.path.join(self._root, sample),
                has_nuc="Nuc" in info["Files"],
                has_mt="Mito" in info["Files"],
            )

            input_files.update(report.input_files())
            self._reports[sample] = report

        output_prefix = os.path.join(self._root, "summary")
        Node.__init__(
            self,
            description="writing summary to %s" % (output_prefix + ".html",),
            input_files=input_files,
            output_files=(output_prefix + ".html", output_prefix + ".css"),
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        with open(os.path.join(temp, "summary.html"), "w") as output_handle:
            menu_entries = self._build_sidemenu()
            html_header = _HTML_HEADER.format(
                Version=paleomix.__version__, MenuEntries=menu_entries
            )
            output_handle.write(html_header)

            for sample in sorted(self._samples):
                self._write_sample_overview(output_handle, sample)

            output_handle.write(_HTML_FOOTER)

    def _teardown(self, config, temp):
        fileutils.make_dirs(self._root)

        fileutils.move_file(
            os.path.join(temp, "summary.html"), os.path.join(self._root, "summary.html")
        )

        css_path = paleomix.resources.report("zonkey", "report.css")
        fileutils.copy_file(css_path, os.path.join(self._root, "summary.css"))

    def _build_sidemenu(self):
        lines = []
        for sample in sorted(self._samples):
            lines.append('        <a href="#sample_{0}">{0}</a>'.format(sample))
        return "\n".join(lines)

    def _write_sample_overview(self, handle, sample):
        info = self._samples[sample]
        sample_idx = self._sample_keys.index(sample)

        admixture = "&nbsp;"
        if "Nuc" in info["Files"]:
            admixture = self._read_admixture_results(sample)

        handle.write(_SAMPLE_HEADER.format(sample=sample, admixture=admixture))

        handle.write("""<ul id="tabs">""")
        handle.write(
            _SAMPLE_TAB_SELECTED.format(sample_idx=sample_idx, page=1, title="Overview")
        )
        handle.write(
            _SAMPLE_TAB.format(sample_idx=sample_idx, page=2, title="Sample Data")
        )

        if "Nuc" in info["Files"]:
            handle.write(
                _SAMPLE_TAB.format(sample_idx=sample_idx, page=3, title="PCA Plots")
            )
            handle.write(
                _SAMPLE_TAB.format(sample_idx=sample_idx, page=4, title="Treemix Plots")
            )

        if "Mito" in info["Files"]:
            handle.write(
                _SAMPLE_TAB.format(
                    sample_idx=sample_idx, page=5, title="Mitochondrial Phylogeny"
                )
            )

        handle.write("""</ul>""")

        self._write_overview(handle, sample, sample_idx)

        if "Nuc" in info["Files"]:
            handle.write(_SAMPLE_PAGE_PCA.format(sample=sample, sample_idx=sample_idx))
            handle.write(
                _SAMPLE_PAGE_TREEMIX.format(sample=sample, sample_idx=sample_idx)
            )

        if "Mito" in info["Files"]:
            handle.write(
                _SAMPLE_PAGE_MITO_PHYLO.format(sample=sample, sample_idx=sample_idx)
            )

        handle.write(_SAMPLE_FOOTER)

    def _write_overview(
        self, output_handle, sample, sample_idx, cutoff=admixture.CUTOFF
    ):
        info = self._samples[sample]
        report = self._reports[sample]

        n_tests = 4
        n_pos_tests = 0

        if "Nuc" in info["Files"]:

            for postfix in ("incl_ts", "excl_ts"):
                admix_root = os.path.join(self._root, sample, "results", "admixture")

                for k_groups in sorted(self._data.groups):
                    filename = os.path.join(admix_root, "%s.%i.Q" % (postfix, k_groups))

                    result = admixture.read_admixture_results(
                        filename, self._data, k_groups
                    )

                    if sum(value >= cutoff for _, value in result) > 1:
                        n_pos_tests += 1

            if n_pos_tests:
                tmpl = _SAMPLE_OVERVIEW_INCL_NUCL_POSITIVE
            else:
                tmpl = _SAMPLE_OVERVIEW_INCL_NUCL_NEGATIVE
        else:
            tmpl = _SAMPLE_OVERVIEW_EXCL_NUCL

        output_handle.write(
            tmpl.format(sample_idx=sample_idx, n_tests=n_tests, n_pos_tests=n_pos_tests)
        )
        output_handle.write(_SAMPLE_OVERVIEW_FOOTER.format(sample=sample))

        output_handle.write(_SAMPLE_DATA_HEADER.format(sample_idx=sample_idx))

        if "Nuc" in info["Files"]:
            summary = report.snp_summary()
            output_handle.write(_SAMPLE_DATA_NUCLEAR % summary)

        if "Mito" in info["Files"]:
            summary = report.mito_summary()
            output_handle.write(_SAMPLE_DATA_MITOCHONDRIA % summary)

        output_handle.write(_SAMPLE_DATA_FOOTER)

    def _read_admixture_results(self, sample, cutoff=admixture.CUTOFF):
        lines = []
        lines.append(
            '<table summary="{} {}" style="width:125px;">'.format(
                "Admixture overview for sample", sample.replace('"', "")
            )
        )
        lines.append("<tr>")
        for postfix in ("incl_ts", "excl_ts"):
            admix_root = os.path.join(self._root, sample, "results", "admixture")

            for k_groups in sorted(self._data.groups):
                filename = os.path.join(admix_root, "%s.%i.Q" % (postfix, k_groups))

                lines.append("<td>")
                try:
                    ancestral_groups = admixture.read_admixture_results(
                        filename, self._data, k_groups, cutoff
                    )
                    lines.extend(
                        self._build_admixture_figure(
                            [value for _, value in ancestral_groups]
                        )
                    )
                except admixture.AdmixtureError:
                    lines.append('<div style="height:100px;background:gray"></div>')
                lines.append("</td>")
        lines.append("</tr>")
        lines.append("</table>\n")

        return "\n".join(lines)

    @classmethod
    def _build_admixture_figure(cls, fractions, max_height=100):
        lines = []
        if len(fractions) == 2:
            colors = ("red", "green")
        elif len(fractions) == 3:
            colors = ("red", "green", "blue")
        else:
            raise RuntimeError("Unexpected number of fractions: %r" % (fractions,))

        values = [value for value in fractions if value >= admixture.CUTOFF]

        tmpl = " " * 18 + '<div style="height:{}px;background:{}"></div>'
        for index, value in enumerate(values):
            height = round((float(value) / sum(values)) * max_height)
            lines.append(tmpl.format(height, colors[index]))

        return lines


###############################################################################

_TS_LABELS = {"incl_ts": "Including transitions", "excl_ts": "Excluding transitions"}


###############################################################################

_HTML_HEADER = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>PALEOMIX Zonkey v{Version}</title>
    <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
    <link href="summary.css" rel="stylesheet" type="text/css" />

    <script type="text/javascript">
        function selectTab(sample, page) {{
            for (var i = 1; i <= 5; ++i) {{
                elem_id = "sample" + sample.toString() + ".link" + i.toString()

                var elem = document.getElementById(elem_id);
                if (elem) {{
                    elem.className = (i == page) ? 'selected' : '';
                }}

                elem_id = "sample" + sample.toString() + ".page" + i.toString()
                var elem = document.getElementById(elem_id);
                if (elem) {{
                    if (i == page) {{
                        if (page == 1) {{
                            elem.className = 'tabContent small';
                        }} else {{
                            elem.className = 'tabContent';
                        }}
                    }} else {{
                        elem.className = 'tabContent hide';
                    }}
                }}
            }}
        }}
    </script>
</head>
<body>
  <a name="top" id="top"></a>
  <center>
    <div id="header">
      <h1>PALEOMIX Zonkey {Version}</h1>
      <h2>A pipeline for detection of F1 hybrids in equids.</h2>
    </div>

    <div id="content" style="width:1050px;">
      <p class="introduction">
        Schubert M, Ermini L, Sarkissian CD, J&oacute;nsson H, Ginolhac A,
        Schaefer R, Martin MD, Fern&aacute;ndez R, Kircher M, McCue M,
        Willerslev E, and Orlando L. "<strong>Characterization of ancient and
        modern genomes by SNP detection and phylogenomic and metagenomic
        analysis using PALEOMIX</strong>". Nat Protoc. 2014 May;9(5):1056-82.
        doi:<a href="https://doi.org/10.1038/nprot.2014.063">
            10.1038/nprot.2014.063
        </a>.
        Epub 2014 Apr 10. PubMed PMID:
        <a href="http://www.ncbi.nlm.nih.gov/pubmed/24722405">24722405</a>.
      </p>
      <div id="sidebar">
        <h1>Contents</h1>
        <div class="submenu">
          <a href="#">Top</a>
          <a href="#intro">Introduction</a>
          {MenuEntries}
        </div>
      </div>
      <div id="mainbar">
        <h1><a name="introduction" id="introduction"></a>Introduction</h1>
        <div>
          <div>
            The Zonkey Pipeline is a easy-to-use pipeline designed for the
            analyses of low-coverage, ancient DNA derived from historical
            equid samples, with the purpose of determining the species of
            the sample, as well as determining possible hybridization between
            horses, zebras, and asses. This is accomplished by comparing one
            or more samples aligned against the <em>Equus caballus</em> 2.0
            reference sequence with a reference panel of modern equids,
            including wild and domesticated equids.
          </div>
          <br/>
          <div>
            For more information, please refer to the
            <a href="http://paleomix.readthedocs.org/en/stable/zonkey_pipeline/index.html">
              the documentation for the Zonkey pipeline
            </a>
            or
            <a href="http://paleomix.readthedocs.org/en/stable/">
              the documentation for the PALEOMIX pipeline,
            </a>
            on which the Zonkey pipeline is based.
          </div>
        </div>
        <h1><a name="samples" id="samples"></a>Sample Overview</h1>
"""  # noqa: E501


_SAMPLE_HEADER = """
    <h2 style="margin: 0px; padding: 0px;">
      <a name="sample_{sample}" id="sample_{sample}"
         href="{sample}/report.html">{sample}</a>
    </h2>

    <table style="margin: 0px; padding: 0px;">
      <tr>
        <td style="width:150px; vertical-align:top; padding-top:20px;">
{admixture}
        </td>
        <td>
"""

_SAMPLE_FOOTER = """
        </td>
      </tr>
    </table>
"""

_SAMPLE_TAB_SELECTED = """
      <li>
        <a id="sample{sample_idx}.link{page}"
           href="javascript:selectTab({sample_idx}, {page});"
           class="selected">{title}</a>
      </li>
"""

_SAMPLE_TAB = """
      <li>
        <a id="sample{sample_idx}.link{page}"
           href="javascript:selectTab({sample_idx}, {page});">{title}</a>
      </li>
"""


_SAMPLE_OVERVIEW_INCL_NUCL_POSITIVE = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <div>
        Possible admixture detected in {n_pos_tests} of {n_tests} tests.
      </div>
"""


_SAMPLE_OVERVIEW_INCL_NUCL_NEGATIVE = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <div>
        No admixture detected in nuclear genome.
      </div>
"""


_SAMPLE_OVERVIEW_EXCL_NUCL = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <div>
        Nuclear BAM not provided; no admixture were tests perfomed.
      </div>
"""


_SAMPLE_OVERVIEW_FOOTER = """
      <div style="text-align:center;">
          <a href="{sample}/report.html">(Click here for the full report)</a>
      </div>
    </div>
"""


_SAMPLE_DATA_HEADER = """
    <div class="tabContent hide" id="sample{sample_idx}.page2">
      <ul>
"""


_SAMPLE_DATA_NUCLEAR = """
      <strong>Nuclear report:</strong>
      <ul>
        <li>BAM file: <em>%(filename)s</em></li>
        <li>Number of SNPs used (incl. transitions):
            <em>%(n_sites_incl_ts)s</em></li>
        <li>Number of SNPs used (excl. transitions):
            <em>%(n_sites_excl_ts)s</em></li>
      </ul>

      <br>
"""


_SAMPLE_DATA_MITOCHONDRIA = """
      <strong>Mitochondrial report:</strong>
      <ul>
        <li>BAM file: <em>%(filename)s</em></li>
        <li>Percentage of sites covered: <em>%(covered_pct)s</em></li>
        <li>Mean coverage per site: <em>%(mean_coverage)s</em></li>
      </ul>
"""


_SAMPLE_DATA_FOOTER = """
      </ul>
    </div>
"""


_SAMPLE_PAGE_PCA = """
    <div class="tabContent hide" id="sample{sample_idx}.page3">
      <table>
        <tr style="background-color:#f1f0ee">
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Including transitions</strong>

              <div class="image">
                <a href="{sample}/figures/pca/incl_ts.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/pca/incl_ts.png"
                       alt="PCA plot for {sample}, including transitions."/>
                </a>
              </div>
            </div>
          </td>
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Excluding transitions</strong>
              <div class="image">
                <a href="{sample}/figures/pca/excl_ts.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/pca/excl_ts.png"
                       alt="PCA plot for {sample}, excluding transitions."/>
                </a>
              </div>
            </div>
          </td>
        </tr>
      </table>
    </div>
"""


_SAMPLE_PAGE_TREEMIX = """
    <div class="tabContent hide" id="sample{sample_idx}.page4">
      <table>
        <tr style="background-color:#f1f0ee">
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Including transitions</strong>
              <div class="image">
                <a href="{sample}/figures/treemix/incl_ts_1_tree.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/treemix/incl_ts_1_tree.png"
                       alt="Treemix plot for {sample}, including transitions,
                            with one migration edge."/>
                </a>
              </div>
            </div>
          </td>
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Excluding transitions</strong>
              <div class="image">
                <a href="{sample}/figures/treemix/excl_ts_1_tree.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/treemix/excl_ts_1_tree.png"
                       alt="Treemix plot for {sample}, excluding transitions,
                            with one migration edge."></img>
                </a>
              </div>
            </div>
          </td>
        </tr>
      </table>
    </div>
"""


_SAMPLE_PAGE_MITO_PHYLO = """
    <div class="tabContent hide" id="sample{sample_idx}.page5">
      <div class="thumbnail" style="text-align: center;">
        <div class="image">
          <a href="{sample}/figures/mitochondria/mito_phylo.pdf">
            <img style="width:15em"
                 src="{sample}/figures/mitochondria/mito_phylo.png"
                 alt="Mitochondrial phylogeny for {sample}."/>
          </a>
        </div>
      </div>
    </div>
"""


_HTML_FOOTER = """
      </div>
    </div>
    <div id="footer">
        This report is based on the PLAIN 1.0 design by
        <a href="http://www.sixshootermedia.com/">6ix Shooter Media</a>,
        Creative Commons license.<br/>
    </div>
  </center>
</body>
</html>
"""



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Version checks for apps or libraries required by PALEOMIX pipelines.

The module contains to sets of classes: RequirementObj and Check. The
RequirementObj class implements the determation of the current version for a
given tool, while the Check (sub)classes implements various comparison
to be carried against the detected version (less than, greater than or equal,
etc.).

To reduce the overhead of detmining versions (which mostly involve invoking
external programs), the RequirementObj caches results. Additionally, to avoid
redundant calls, RequirementObjs are created using the 'Requirement' function
which caches RequirementObjs.

For example, to check that the Java version is v1.7 or later:
    obj = Requirement(call=("java", "-version"),
                      search='java version "(\\d+).(\\d+)',
                      checks=GE(1, 7),
                      name="Java Runtime Environment")
    try:
        obj()
    except VersionRequirementError:
        pass  # requirements not met, or failure to determine version
"""
import operator
import re

from paleomix.common.utilities import TotallyOrdered, safe_coerce_to_tuple, try_cast

import paleomix.common.procs as procs


# Cache used to store the output of cmd-line / function calls
_CALL_CACHE = {}
# Cache used to store Requirement object
_REQUIREMENT_CACHE = {}


class VersionRequirementError(Exception):
    """Raised if version requirements are not met, or if a version could not be
    determined for a requirement check.
    """


def Requirement(call, search, checks, name=None, priority=0):
    # Ignore function naming scheme
    """Returns a singleton Requirement object, based on the parameters,
    which may be used to check that version requirements are met for a
    given program/utility/module, etc.

    Parameters:
      call   -- A string, or a tuple containing strings for a system call,
                or a tuple containing a function at the first position, and
                a set of positional parameters. In the case of system calls,
                stdout and stderr are returned as a single string, in the case
                of a function call, the return value is expected to be a str.
      search -- A regular expression (string or re object), used to search
                the output of the "call". Groups are assumed to represent
                version numbers.
      checks -- A callable that implements the interface described in the
                Check class.
      name   -- Descriptive name for the executable/module/etc. If not
                specified, first value in 'call' will be used; if multiple
                otherwise identical checks are made, the last name that
                does not equal the first value of 'call' will be used.
      priority -- Order in which requirements are checked; if multiple
                  otherwise identical checks are made with different priority,
                  the highest priority takes precedence.

    Implementation detail: To reduce the need for performing calls or system-
    calls multiple times, caches are implemented using the call object as keys.
    Thus the same calls should be passed in a manner which allow equality
    between the same calls to be established.
    """
    call = safe_coerce_to_tuple(call)
    key = (call, search, checks)

    try:
        requirement = _REQUIREMENT_CACHE[key]

        # Highest priority takes precedence
        requirement.priority = max(requirement.priority, priority)
        # Last explicitly specified name takes precedence
        requirement.name = name or requirement.name
    except KeyError:
        requirement = RequirementObj(*key, name=name, priority=priority)
        _REQUIREMENT_CACHE[key] = requirement

    return requirement


class RequirementObj:
    """Represents a version requirement."""

    def __init__(self, call, search, checks, name=None, priority=0):
        """See function 'Requrement' for a description of parameters.
        """
        self._call = safe_coerce_to_tuple(call)
        self._done = None
        self.name = str(name or self._call[0])
        self.priority = int(priority)
        self.checks = checks
        self._rege = re.compile(search)
        self._version = None

    @property
    def version(self):
        """The version determined for the application / library. If the version
        could not be determined, a VersionRequirementError is raised,
        describing the cause of the problem.
        """
        if self._version is None:
            try:
                output = _do_call(self._call)
            except OSError as error:
                self._raise_failure(error)

            if isinstance(output, bytes):
                output = output.decode("utf-8", "replace")

            # Raise an exception if the JRE is outdated, even if the
            # version could be determined (likely a false positive match).
            self._check_for_outdated_jre(output)

            match = self._rege.search(output)
            if not match:
                self._raise_failure(output)

            self._version = tuple(
                0 if value is None else try_cast(value, int) for value in match.groups()
            )

        return self._version

    @property
    def executable(self):
        """Returns the executable invoked during version determination; if no
        executable is invoked, None is returned.
        """
        if not callable(self._call[0]):
            return self._call[0]

    def __call__(self, force=False):
        if force or self._done is None:
            if not self.checks(self.version):
                raise VersionRequirementError(
                    "Version requirements not met for %s: Expected %s, but found %s"
                    % (self.name, self.checks, _pprint_version(self.version))
                )

            self._done = True

    def _check_for_outdated_jre(self, output):
        """Checks for the error raised if the JRE is unable to run a JAR file.
        This happens if the JAR was built with a never version of the JRE, e.g.
        if Picard was built with a v1.7 JRE, and then run with a v1.6 JRE.
        """
        # This exception is raised if the JRE is incompatible with the JAR
        if "UnsupportedClassVersionError" in output:
            self._raise_failure(output)

    def _raise_failure(self, output):
        """Raises a VersionRequirementError when a version check failed; if the
        output indicates that the JRE is outdated (i.e. the output contains
        "UnsupportedClassVersionError") a special message is givenself.
        """
        lines = ["Version could not be determined for %s:" % (self.name,)]
        lines.extend(self._describe_call())
        lines.append("")

        if isinstance(output, OSError):
            lines.append("Exception was raised:")
            lines.append("    %s: %s" % (output.__class__.__name__, output))
        elif "UnsupportedClassVersionError" in output:
            # Raised if the JRE is too old compared to the JAR
            lines.extend(
                [
                    "The version of the Java Runtime Environment on this",
                    "system is too old; please check the the requirement",
                    "for the program and upgrade your version of Java.",
                    "",
                    "See the documentation for more information.",
                ]
            )
        else:
            lines.append("Program may be broken or a version not supported by the")
            lines.append("pipeline; please refer to the PALEOMIX documentation.\n")
            lines.append("    Required:       %s" % (self.checks,))
            lines.append("    Search string:  %s\n" % (self._rege.pattern))
            lines.append("%s Command output %s" % ("-" * 22, "-" * 22))
            lines.append(output)

        raise VersionRequirementError("\n".join(lines))

    def _describe_call(self):
        """Returns lines describing the current system call, if any."""
        if not callable(self._call[0]):
            yield "    Command  = %s" % (" ".join(self._call),)


class Check(TotallyOrdered):
    """Abstract base-class for version checks.

    Callable with a tuple of version fields (typically integers), and returns
    either True or False, depending on whether or not the specified check
    passed.

    The contructor takes a string describing the check ('description'), a
    function with the signature 'func(version, values)', where version is the
    version determined for a app/library, and where values are the values
    passed to the Check constructor.
    """

    def __init__(self, description, func, *values):
        if not callable(func):
            raise TypeError("func must be callable, not %r" % (func,))

        values = tuple(values)
        self._func = func
        self._values = values
        self._description = description
        self._objs = (description, func, values)

    def __str__(self):
        return self._description

    def __eq__(self, other):
        if isinstance(other, Check):
            return self._objs == other._objs
        return NotImplemented

    def __hash__(self):
        return hash(self._objs)

    def __call__(self, current):
        """Takes a tuple of version fields (e.g. (1, 7)) and returns True if
        this version matches the check.
        """
        return self._do_check_version(current, self._values)

    def _do_check_version(self, current, reference):
        """Invokes the actual check; may be overridden in subclasses."""
        return self._func(current, reference)


class CheckVersion(Check):
    """Base class for comparisons involving versions; requires that the version
    checks has at least as many fields as specified for the Check object. If
    the version checked has more fields, these are truncated away.
    """

    def __init__(self, description, func, *version):
        description = description.format(_pprint_version(version))
        Check.__init__(self, description, func, *version)

    def _do_check_version(self, current, reference):
        if len(current) < len(reference):
            raise ValueError(
                "Expects at least %i fields, not %i: %r"
                % (len(reference), len(current), current)
            )

        return Check._do_check_version(self, current[: len(reference)], reference)


class EQ(CheckVersion):
    """Checks that a version is Equal to this version; note that version fields
    are truncated to the number of fields specified for this Check. As a
    consequence, EQ(1, 5) is true for (1, 5), (1, 5, 7), (1, 5, 7, 1), etc. See
    'Check' for more information.
    """

    def __init__(self, *version):
        CheckVersion.__init__(self, "{0}", operator.eq, *version)


class GE(CheckVersion):
    """Checks that a version is Greater-than or Equal to this version; note
    that version fields are truncated to the number of fields specified for
    this Check. See 'Check'.
    """

    def __init__(self, *version):
        CheckVersion.__init__(self, "at least {0}", operator.ge, *version)


class LT(CheckVersion):
    """Checks that a version is Less Than this version; note that version
    fields are truncated to the number of fields specified for this Check.
    See 'Check'.
    """

    def __init__(self, *version):
        CheckVersion.__init__(self, "prior to {0}", operator.lt, *version)


class Any(CheckVersion):
    """Dummy check; is always true."""

    def __init__(self):
        CheckVersion.__init__(self, "any version", _func_any)


class Operator(Check):
    """Base class for logical operations on Checks; and, or, etc."""

    def __init__(self, keyword, func, *checks):
        """Arguments:
          keyword -- Keyword to join description of checks by.
          func -- Function implementing the logical operation; is called as
                  func(*checks). See the 'func' argument for Check.__init__.
          checks -- Zero or more Checks.
        """
        descriptions = []
        for check in checks:
            if isinstance(check, Operator):
                descriptions.append("(%s)" % (check,))
            elif isinstance(check, Check):
                descriptions.append("%s" % (check,))
            else:
                raise ValueError("%r is not of type Check" % (check,))

        description = keyword.join(descriptions)
        Check.__init__(self, description, func, *checks)


class And(Operator):
    """Carries out 'and' on a set of checks; always true for no Checks"""

    def __init__(self, *checks):
        Operator.__init__(self, " and ", _func_and, *checks)


class Or(Operator):
    """Carries out 'or' on a set of checks; always false for no Checks"""

    def __init__(self, *checks):
        Operator.__init__(self, " or ", _func_or, *checks)


###############################################################################
###############################################################################
# Check functions; must be available for pickle


def _func_any(_current, _checks):
    """Implementation of Any."""
    return True


def _func_and(current, checks):
    """Implementation of And."""
    return all(check(current) for check in checks)


def _func_or(current, checks):
    """Implementation of Or."""
    return any(check(current) for check in checks)


###############################################################################
###############################################################################
# Utility functions


def _run(call):
    """Carries out a system call and returns STDOUT and STDERR as a combined
    string. If an OSError is raied (e.g. due to missing executables), the
    resulting message is returned as a string. If the call raised an OSError,
    then the exception is returned as a value.
    """
    try:
        proc = procs.open_proc(
            call,
            stdout=procs.PIPE,
            # Merge STDERR with STDOUT output
            stderr=procs.STDOUT,
        )

        return proc.communicate()[0]
    except OSError as error:
        return error


def _do_call(call):
    """Performs a call; the result is cached, and returned upon subsequent
    calls with the same signature (either a function call or system call). If
    the call raised an OSError, then the exception is returned as a value.
    """
    try:
        result = _CALL_CACHE[call]
    except KeyError:
        if callable(call[0]):
            result = call[0](*call[1:])
        else:
            result = _run(call)
        _CALL_CACHE[call] = result

    if isinstance(result, OSError):
        raise result

    return result


def _pprint_version(value):
    """Pretty-print version tuple; takes a tuple of field numbers / values,
    and returns it as a string joined by dots with a 'v' prepended.
    """
    return "v%s" % (".".join(map(str, value)),)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.node import Node, NodeError

from paleomix.common.bedtools import read_bed_file, pad_bed_records, merge_bed_records
from paleomix.common.fileutils import move_file, reroot_path


class PaddedBedNode(Node):
    """Simple node for padding BED records a fixed amount and merging
    overlapping records. Columns beyond the 3rd column are dropped.
    """

    def __init__(self, infile, outfile, fai_file, amount=0, dependencies=()):
        self._amount = int(amount)
        self._infile = infile
        self._outfile = outfile
        self._fai_file = fai_file

        Node.__init__(
            self,
            description="padding BED records (%i) in %s" % (amount, infile),
            input_files=(infile, fai_file),
            output_files=(outfile,),
            dependencies=dependencies,
        )

    def _run(self, config, temp):
        contigs = {}
        with open(self._fai_file) as handle:
            for line in handle:
                name, length, _ = line.split("\t", 2)
                if name in contigs:
                    raise NodeError(
                        "Reference genome contains multiple "
                        "identically named contigs (%r)!" % (name,)
                    )

                contigs[name] = int(length)

        with open(reroot_path(temp, self._outfile), "w") as handle:
            records = list(read_bed_file(self._infile, contigs=contigs))
            pad_bed_records(records=records, padding=self._amount, max_sizes=contigs)

            for record in merge_bed_records(records):
                handle.write("%s\n" % (record,))

    def _teardown(self, config, temp):
        source = reroot_path(temp, self._outfile)
        move_file(source, self._outfile)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os

import paleomix.common.logging
import paleomix.pipelines.phylo.example as example
import paleomix.pipelines.phylo.mkfile as mkfile
import paleomix.pipelines.phylo.parts.genotype as genotype
import paleomix.pipelines.phylo.parts.msa as msa
import paleomix.pipelines.phylo.parts.phylo as phylo
import paleomix.yaml

from paleomix.pipeline import Pypeline
from paleomix.pipelines.phylo.config import build_parser
from paleomix.pipelines.phylo.makefile import MakefileError, read_makefiles


_COMMANDS = {
    "genotype": genotype.chain,
    "msa": msa.chain,
    "phylogeny": phylo.chain_examl,
}


def main(argv):
    log = logging.getLogger(__name__)

    commands = [it.lower().strip() for it in argv[0].split("+")] if argv else []
    if "example" in commands:
        return example.main(argv[1:])

    parser = build_parser()
    config = parser.parse_args(argv)
    config.commands = commands
    if "help" in config.commands:
        parser.print_help()
        return 0
    elif "makefile" in config.commands or "mkfile" in config.commands:
        return mkfile.main(config.files)

    commands = []
    for key in config.commands:
        func = _COMMANDS.get(key)
        if func is None:
            log.error("unknown command %r", key)
            return 1

        commands.append((key, func))

    paleomix.common.logging.initialize(
        log_level=config.log_level,
        log_file=config.log_file,
        name="phylo_pipeline",
    )

    if not os.path.exists(config.temp_root):
        try:
            os.makedirs(config.temp_root)
        except OSError as error:
            log.error("Could not create temp root:\n\t%s", error)
            return 1

    if not os.access(config.temp_root, os.R_OK | os.W_OK | os.X_OK):
        log.error("Insufficient permissions for temp root: %r", config.temp_root)
        return 1

    # Init worker-threads before reading in any more data
    pipeline = Pypeline(config)

    try:
        makefiles = read_makefiles(config, commands)
    except (MakefileError, paleomix.yaml.YAMLError, IOError) as error:
        log.error("Error reading makefiles:\n%s", error)
        return 1

    for (command_key, command_func) in commands:
        log.info("Building %s pipeline", command_key)
        command_func(pipeline, config, makefiles)

    for makefile in makefiles:
        if "Nodes" in makefile:
            pipeline.add_nodes(makefile["Nodes"])

    if config.list_input_files:
        log.info("Printing output files")
        pipeline.print_input_files()
        return 0
    elif config.list_output_files:
        log.info("Printing output files")
        pipeline.print_output_files()
        return 0
    elif config.list_executables:
        log.info("Printing required executables")
        pipeline.print_required_executables()
        return 0

    if not pipeline.run(max_threads=config.max_threads, dry_run=config.dry_run):
        return 1

    return 0



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bz2
import gzip
import io

import pytest

from paleomix.common.fileutils import fspath
from paleomix.common.formats.fastq import FASTQ, FASTQError, FASTQualities


_SEQ_FRAG = "AAGTCC"  # len() = 6
_QUAL_FRAG = "123456"  # len() = 6


###############################################################################
###############################################################################
# Tests for FASTQ constructor


def _simple_fastq_record():
    return FASTQ("Dummy", "Meta-inf", "ACGT", "1234")


def test_fastq__constructor():
    record = _simple_fastq_record()
    assert record.name == "Dummy"
    assert record.meta == "Meta-inf"
    assert record.sequence == "ACGT"
    assert record.qualities == "1234"


def test_fastq__constructor__no_meta():
    record = FASTQ("Dummy", None, "ACGT", "1234")
    assert record.name == "Dummy"
    assert record.meta == ""
    assert record.sequence == "ACGT"
    assert record.qualities == "1234"


def test_fastq__constructor__name_must_be_string_type():
    with pytest.raises(FASTQError, match="FASTQ name must be a non-empty string"):
        FASTQ(1, None, "ACGT", "1234")


def test_fastq__constructor__name_must_be_non_empty():
    with pytest.raises(FASTQError, match="FASTQ name must be a non-empty string"):
        FASTQ("", None, "ACGT", "1234")


def test_fastq__constructor__meta_must_be_string_type_or_none():
    with pytest.raises(FASTQError, match="FASTQ meta must be a string, or None"):
        FASTQ("Seq1", 1, "ACGT", "1234")


def test_fastq__constructor__seq_must_be_string_type():
    with pytest.raises(FASTQError, match="FASTQ sequence must be a string"):
        FASTQ("Seq1", None, 1, "1234")


def test_fastq__constructor__qual_must_be_string_type():
    with pytest.raises(FASTQError, match="FASTQ qualities must be a string"):
        FASTQ("Seq1", None, "ACGT", 1234)


def test_fastq__constructor__seq_and_qual_must_have_same_len():
    with pytest.raises(FASTQError, match="FASTQ sequence and qualities length differ"):
        FASTQ("Seq1", None, "ACGT", "12345")


###############################################################################
###############################################################################
# Tests for __repr__


def test_fastq__repr():
    expected = "FASTQ('foobar', '', %r, %r)" % (_SEQ_FRAG, _QUAL_FRAG)
    result = repr(FASTQ("foobar", None, _SEQ_FRAG, _QUAL_FRAG))
    assert result == expected


def test_fastq__repr_with_meta_information():
    expected = "FASTQ('foobar', 'my Meta-Info', %r, %r)" % (_SEQ_FRAG, _QUAL_FRAG)
    result = repr(FASTQ("foobar", "my Meta-Info", _SEQ_FRAG, _QUAL_FRAG))
    assert result == expected


###############################################################################
###############################################################################
# Tests for write


def test_fastq__write():
    expected = "@foobar\n%s\n+\n%s\n" % (_SEQ_FRAG, _QUAL_FRAG)
    stringf = io.StringIO()
    FASTQ("foobar", None, _SEQ_FRAG, _QUAL_FRAG).write(stringf)
    assert stringf.getvalue() == expected


def test_fastq__write_with_meta_information():
    expected = "@foobar my Meta-Info\n%s\n+\n%s\n" % (_SEQ_FRAG, _QUAL_FRAG)
    stringf = io.StringIO()
    FASTQ("foobar", "my Meta-Info", _SEQ_FRAG, _QUAL_FRAG).write(stringf)
    assert stringf.getvalue() == expected


###############################################################################
###############################################################################
# Tests for FASTQ.from_lines


def test_fasta__from_lines__no_records():
    assert list(FASTQ.from_lines([])) == list([])


def test_fasta__from_lines__no_records__empty_line():
    assert list(FASTQ.from_lines([""])) == list([])


def test_fasta__from_lines_single_record():
    lines = ["@single\n", "CCTTCATCCA\n", "+", "1234567890"]
    expected = [FASTQ("single", None, "CCTTCATCCA", "1234567890")]
    assert list(FASTQ.from_lines(lines)) == list(expected)


def test_fasta__from_lines__multiple_records():
    lines = [
        "@first\n",
        "TGTTCTCCACCGTGCACAAC\n",
        "+",
        "12345678901234567890\n",
        "@Second XT:1:0\n",
        "GAGAGCTCAGCTAAC\n",
        "+\n",
        "098765432109876\n",
        "@Third\n",
        "GGCATTCGGC\n",
        "+\n",
        "5678901234\n",
    ]
    expected = [
        FASTQ("first", None, "TGTTCTCCACCGTGCACAAC", "12345678901234567890"),
        FASTQ("Second", "XT:1:0", "GAGAGCTCAGCTAAC", "098765432109876"),
        FASTQ("Third", None, "GGCATTCGGC", "5678901234"),
    ]
    assert list(FASTQ.from_lines(lines)) == list(expected)


def test_fasta__from_lines__partial__1_line():
    with pytest.raises(FASTQError, match="Partial FASTQ record"):
        list(FASTQ.from_lines(["@fastq1\n"]))


def test_fasta__from_lines__partial__2_lines():
    with pytest.raises(FASTQError, match="Partial FASTQ record"):
        list(FASTQ.from_lines(["@fastq1\n", "ACGT\n"]))


def test_fasta__from_lines__partial__3_lines():
    with pytest.raises(FASTQError, match="Partial FASTQ record"):
        list(FASTQ.from_lines(["@fastq1\n", "ACGT\n", "+\n"]))


def test_fasta__from_lines__invalid_header():
    lines = [">fastq\n", "GGCATTCGGC\n", "+\n", "5678901234\n"]
    with pytest.raises(FASTQError, match="Invalid FASTQ header"):
        list(FASTQ.from_lines(lines))


def test_fasta__from_lines__mismatching_lengths():
    lines = ["@fastq\n", "GGCATTCGGC\n", "+\n", "567890123\n"]
    with pytest.raises(
        FASTQError, match="Sequence length does not match qualities length"
    ):
        list(FASTQ.from_lines(lines))


def test_fasta__from_lines__invalid_separator():
    lines = ["@fastq\n", "GGCATTCGGC\n", "?\n", "5678901234\n"]
    with pytest.raises(FASTQError, match="Invalid FASTQ separator"):
        list(FASTQ.from_lines(lines))


###############################################################################
###############################################################################
# Tests for 'FASTQ.from_file'


@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_fasta__from_file(func, tmp_path):
    expected = [
        FASTQ("This_is_FASTA!", None, "ACGTN", "12345"),
        FASTQ("This_is_ALSO_FASTA!", None, "CGTNA", "56789"),
    ]

    with func(fspath(tmp_path / "file"), "wt") as handle:
        for item in expected:
            item.write(handle)

    assert list(FASTQ.from_file(tmp_path / "file")) == expected


###############################################################################
###############################################################################
# Tests for 'FASTQualities'

_33_READ = FASTQ("33", None, "ACGT", "!02I")
_64_READ = FASTQ("33", None, "ACGT", "@UXi")
_AMBIGIOUS_read = FASTQ("33", None, "ACGT", ";CDI")


def test_fastqualities__no_qualities():
    quals = FASTQualities()

    assert quals.offsets() == FASTQualities.MISSING


def test_fastqualities__phred_33():
    quals = FASTQualities()
    quals.update(_33_READ)

    assert quals.offsets() == FASTQualities.OFFSET_33


def test_fastqualities__phred_64():
    quals = FASTQualities()
    quals.update(_64_READ)

    assert quals.offsets() == FASTQualities.OFFSET_64


def test_fastqualities__phred_ambigious():
    quals = FASTQualities()
    quals.update(_AMBIGIOUS_read)

    assert quals.offsets() == FASTQualities.AMBIGIOUS


def test_fastqualities__phred_both():
    quals = FASTQualities()
    quals.update(_33_READ)
    quals.update(_64_READ)

    assert quals.offsets() == FASTQualities.BOTH


###############################################################################
###############################################################################


def test_fasta__equality():
    assert FASTQ("A", "B", "C", "D") == FASTQ("A", "B", "C", "D")


def test_fasta__inequality():
    assert FASTQ("A", "B", "C", "D") != FASTQ("A", "B", "D", "D")
    assert FASTQ("A", "B", "C", "D") != FASTQ("A", None, "C", "D")
    assert FASTQ("A", "B", "C", "D") != FASTQ("D", "B", "C", "D")


def test_fasta__sorting_less_equal():
    assert not FASTQ("A", "B", "C", "D") < FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "C", "D") < FASTQ("B", "B", "C", "D")
    assert FASTQ("A", "B", "C", "D") < FASTQ("A", "C", "C", "D")
    assert FASTQ("A", "B", "C", "D") < FASTQ("A", "B", "D", "D")
    assert FASTQ("A", "B", "C", "D") <= FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "C", "D") <= FASTQ("B", "B", "C", "D")
    assert FASTQ("A", "B", "C", "D") <= FASTQ("A", "C", "C", "D")
    assert FASTQ("A", "B", "C", "D") <= FASTQ("A", "B", "D", "D")


def test_fasta__sorting_greater_equal():
    assert not FASTQ("A", "B", "C", "D") > FASTQ("A", "B", "C", "D")
    assert FASTQ("B", "B", "C", "D") > FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "C", "C", "D") > FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "D", "D") > FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "C", "E") > FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "C", "D") >= FASTQ("A", "B", "C", "D")
    assert FASTQ("B", "B", "C", "D") >= FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "C", "C", "D") >= FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "D", "D") >= FASTQ("A", "B", "C", "D")
    assert FASTQ("A", "B", "C", "E") >= FASTQ("A", "B", "C", "D")


def test_fasta__hash():
    assert hash(FASTQ("A", "B", "C", "D")) == hash(FASTQ("A", "B", "C", "D"))
    assert hash(FASTQ("A", "B", "C", "D")) != hash(FASTQ("B", "B", "C", "D"))
    assert hash(FASTQ("A", "B", "C", "D")) != hash(FASTQ("A", "C", "C", "D"))
    assert hash(FASTQ("A", "B", "C", "D")) != hash(FASTQ("A", "B", "D", "D"))
    assert hash(FASTQ("A", "B", "C", "D")) != hash(FASTQ("A", "B", "C", "E"))


def test_fasta__unimplemented_comparison():
    assert NotImplemented is FASTQ("A", None, "C", "D").__eq__(10)
    assert NotImplemented is FASTQ("A", None, "C", "D").__lt__(10)
    assert NotImplemented is FASTQ("A", None, "C", "D").__le__(10)
    assert NotImplemented is FASTQ("A", None, "C", "D").__ge__(10)
    assert NotImplemented is FASTQ("A", None, "C", "D").__gt__(10)
